/*
 * Copyright 2017 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"

#include "CoZPacket.h"
#include "CoZClient.h"
#include "CoZUtil.h"

#define Str(x) #x
#define Xstr(x) Str(x)
const char* COZ_VERSION = Xstr(BUILD_VERSION) ;

/**
 * CoZControl-cmd - command line client for the cozcontrol command.
 *
 * All of the heavy lifting is done by CozClient class.
 * 
 */

void usage() {

	fprintf(stderr,
			"usage: cozcontrol start [-t] -ssh [ssh-opt*] [user]@host  \n"
			"   or: cozcontrol stop \n"
			"   or: cozcontrol -h\n"
			"   or: cozcontrol -v\n"
			"\n"
			"options:\n"
			"  -h  -> display this help and exit\n"
			"  -t  -> configure a socket connection, forwarded over a control master SSH connection\n"
			"  -v  -> display the current version and exit\n"
	);

}

int main(int argc, char** argv) {

	int i;
	int ssh_opt_start = 0;
	int ssh_opt_num = 0;
	bool tunneling = false;

	const char* cmd_name = "cozcontrol";
	char pname[50];
	Stringcopy(pname, cmd_name, sizeof(pname));
	SNPRINTF(pname+strlen(pname), sizeof(pname)-strlen(pname), "(%d)", getpid());
	DtlLog::setProgName(pname);
	DtlLog* log = DtlLog::getLogNamed("CoZClient");
	log->printLogName(false);
	log->logmsg(LOG_INFO, "version: %s", COZ_VERSION);

	CoZControlClient client;

	char * p_coz_ssh_opts = getenv(ENV_COZ_SSH_OPTS);

	int arg_cmd_offset = -1;

	int return_code;
	try {

		for (i=1; i<argc; i++) {
			if (!strcmp(argv[i], "-h")) {
				usage();
	   	 		exit(RC_OK);
			} else if (!strcmp(argv[i], "-v")) {
				printf("%s\n", COZ_VERSION);
    			exit(RC_OK);
			} else if (!strcmp(argv[i], "start")) {
				arg_cmd_offset = i;
				i++;
				if (i<argc && !strcmp(argv[i], "-t")) {
					// Note: tunneling is not supported under CYGWIN on Windows
					tunneling = true;
					i++;
				}
			    if (i<argc && !strcmp(argv[i], "-ssh")) {
				    // Mark the start and number of ssh options
					for (ssh_opt_start=++i; i<argc; i++) {
						if (strchr(argv[i], '@')) {
							ssh_opt_num = i - ssh_opt_start + 1;
							break;
						}
					}
					if (ssh_opt_num) {
							break; // done; got ssh options, user@host, start
					} else {
						LOG(LOG_ERR, "-ssh options invalid; must end with user@host\n");
						usage();
						exit(RC_INPUT_ERROR);
					}
			    } else  {
			    	if (!p_coz_ssh_opts) {
						LOG(LOG_ERR, "-ssh option is required with start\n");
						usage();
						exit(RC_INPUT_ERROR);
			        }
			    }
			} else if (!strcmp(argv[i], "stop")) {
				arg_cmd_offset = i;
				break; // done; got stop command
			} else {
				usage();
	   	 		exit(RC_OK);
			}
		}

		int arg_offset = i;

		if (argc - arg_offset == 0) {
			usage();
			exit(RC_INPUT_ERROR);
		}

		if (ssh_opt_start) {
			client.setSshOptions(argv, ssh_opt_start, ssh_opt_num, tunneling);
        } else if (p_coz_ssh_opts) {
			// If the COZ_SSH_OPTS environment variable was set, then we use the given
			// options after the ENV_SSH_CMD, and force an ssh type connection.
			char* ssh_opt_string = new char[strlen(p_coz_ssh_opts)+1];
			strcpy(ssh_opt_string,p_coz_ssh_opts);
			int num_ssh_opts = 50;
			char** ssh_opts = new char*[num_ssh_opts+1];
			CoZUtil::parseCommand(ssh_opt_string, ssh_opts, num_ssh_opts);
			client.setSshOptions(ssh_opts, 0, num_ssh_opts, tunneling);
			delete [] ssh_opts;  // don't need pointers, but need to keep ssh_opt_string
		}

		client.setIOConfigStdin(IO_CONFIG_NONE);
		client.setIOConfigStdout(IO_CONFIG_NONE);
		client.setIOConfigStderr(IO_CONFIG_TEXT);

		return_code = client.exec(cmd_name, argv, arg_cmd_offset, argc-arg_cmd_offset); // user arg_cmd_offset to ensure the start or stop is passed

	} catch (const int rc) {
		return_code = rc;
	} catch (...) {
		LOG(LOG_ERR, "caught unexpectd exception");
    	return_code = RC_SYSTEM_ERROR;
	}

	return return_code;
}

