/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#include "DtlLog.h"

#define Str(x) #x
#define Xstr(x) Str(x)
#ifdef ENV_PREFIX
const char* LOG_PREFIX = Xstr(ENV_PREFIX);
#else
const char* LOG_PREFIX = "DTL";
#endif

extern char **environ;

const int DtlLogConfig::LOG_TABLE_MAX_SIZE = 50;
const int DtlLog::LOG_BUF_MAX = 1024;

DtlLog* DtlLog::_DEFAULT_LOG = new DtlLog();

DtlLogConfig::DtlLogConfig() :
	elapsed_time(NULL),
	log_func(NULL),
	aux_log_func(NULL),
	call_aux_log_func(false),
	log_stream(stderr),
	log_table_size(0),
	log_to_syslog(false),
	print_time(false),
	prog_name(NULL)
{
	log_table = new LogTableEntry[LOG_TABLE_MAX_SIZE];
	char* env_log_stream = new char[strlen(LOG_PREFIX) + 50];
	strcpy(env_log_stream, LOG_PREFIX);
	strcat(env_log_stream, "_DEFAULT_LOGSTREAM");
	char* p = getenv(env_log_stream);
	if (p && strlen(p) > 0) {
		char* s = new char[strlen(p) + 1];
		strcpy(s, p);
		if (FILE* l = fopen(s, "w")) log_stream = l;
		delete[] s;
	}
	delete[] env_log_stream;
}

DtlLog::DtlLog() :
	print_log_name(false),
	level(LOG_NOTICE)
{
	// initialize an unnamed (the default) log
	_DEFAULT_LOG = this;  // to ensure initialization first time through
	char* env_log = new char[strlen(LOG_PREFIX) + 5];
	strcpy(env_log, LOG_PREFIX);
	strcat(env_log, "_LOG");
	char* p = getenv(env_log);
	if (p) {
		char* s = new char[strlen(p) + 1];
		strcpy(s, p);
		handleLogOption(s);
		delete[] s;
	}
	delete[] env_log;
}

/**
 * Construct a named log and save the instance in the log_table
 */
DtlLog::DtlLog(const char* _log_name, int init_level) :
	print_log_name(true),
	level(init_level)
{
	DtlLogConfig& config = getConfig();
	// initialize a named log
	log_name = new char[strlen(_log_name)+1];
	strcpy(log_name, _log_name);
	for (int i=0; i< config.log_table_size; i++) {
		if (!strcmp(log_name, config.log_table[i].name)) {
			DtlLog* prev = config.log_table[i].log;
			level = prev->level;
			config.log_table[i].log = this;
			return;
		}
	}
	if (config.log_table_size == config.LOG_TABLE_MAX_SIZE) {
		fprintf(stderr, "DtlLog: LOG_TABLE_MAX_SIZE exceeded");
		exit(102);
	}
	int i = config.log_table_size++;
	config.log_table[i].log = this;
	config.log_table[i].name = log_name;
}

DtlLog::~DtlLog() {
	if (log_name) delete[] log_name;
}

/**
 * Look up the DtlLog with a given name; answer a new log if not found
 */
DtlLog* DtlLog::getLogNamed(const char* _log_name) {
	DtlLogConfig& config = getConfig();
	for (int i=0; i < config.log_table_size; i++) {
		if (!strcasecmp(_log_name, config.log_table[i].name)) {
			return config.log_table[i].log;
		}
	}
	return new DtlLog(_log_name);
}

/*
 * Handle the various logging options sent via the logging option command switch,
 * or COZ_LOG environment variable.
 *
 * The following options are available:
 * 	'[!]e'  - Show elapsed timestamps on messages
 *  '[!]t'  - Print time [hh:mm:ss.mmm] on messages (default is on)
 *  '[!]s'  - Messages are logged to syslog()
 *  [M|A|C|E|W|N|I|D|T|F] - Set default logging to log at level
 *  f=<logfile> - Set the default logstream to logfile
 *  logname=[M|A|C|E|W|N|I|D|T|F] - Set the named log to log at level
 *
 * Several option settings can be set at once by separating them with either
 * commas or spaces.
 *
 */
void DtlLog::handleLogOption(const char* opt_list) {

	DtlLog* defaultLog = getDefaultLog();
	DtlLogConfig& config = getConfig();
	char* optlist = new char[strlen(opt_list)+1];
	strcpy(optlist, opt_list);
	char* endp;
	char* opt = strtok_r(optlist, ", ", &endp);
	while (opt) {
		int len = strlen(opt);
		bool onoff = true;
		if (len == 2 && *opt == '!') {
			onoff = false;
			len--;
			opt++;
		}
		if (len == 1 ) {
			if (*opt == 'e') {
				//-Le
				defaultLog->setPrintElapsedTime(onoff);
			} else if (*opt == 't') {
				defaultLog->setPrintTime(onoff);
			} else if (opt[0] == 's') {
				//-Ls
				if (onoff) {
					closelog();
					openlog(config.prog_name ? config.prog_name : LOG_PREFIX , LOG_CONS & LOG_PID, LOG_USER);
					setLogFunc(&DtlLog::syslog_writer);
					config.log_to_syslog = true;
				} else {
					closelog();
					setLogFunc(NULL);
					config.log_to_syslog = false;
				}
			} else {
				//-L[MACEWNIDT]
				defaultLog->setLevel(opt);
			}
		} else if (strchr(opt, '=')) {
			if ((opt[0] == 'f' || opt[0] == 'F') && opt[1] == '=') {
				//-Lf=<logfile>
				FILE* f;
				if (!(f = fopen(opt+2, "a"))) {
					LOG(LOG_ERR,"unable to open logstream file: %s - (%d) %s", opt+2, errno, strerror(errno));
				} else {
					defaultLog->setLogStream(f);
				}
			} else {
				//-Llogname=level
				char* lll;
				char* logname = strtok_r(opt, "=", &lll);
				char* level = strtok_r(NULL, "", &lll);
				DtlLog* log = getLogNamed(logname);
				log->setLevel(level);
			}
		} else {
			LOG(LOG_ERR,"Invalid logging option: %s",opt);
		}

		opt = strtok_r(NULL, ", ", &endp);

	} // end of while(opt)
	delete[] optlist;
}

/**
 * Set the level of this log using one of the LOG_SEVERITY_CODES, or an integer level
 */
void DtlLog::setLevel(const char* newLevel) {
	if (*newLevel == '\0') {
		level = -1;  // if empty, then delegate to _DEFAULT_LOG->level
		return;
	}
    if (isdigit(*newLevel)) {
    	level = atoi(newLevel);
    	return;
    }
	int i;
	for (i=0; i<strlen(LOG_SEVERITY_CODES); i++) {
		if(*newLevel == LOG_SEVERITY_CODES[i]) {
			level = i;
			return;
		}
	}
	LOG(LOG_WARNING, "Invalid logging level '%s'.  Valid values are: '%s'", newLevel, LOG_SEVERITY_CODES);
}

/**
 * Log a message if the log level is <= the receivers level.
 * The message is given by a format string and arguments.
 *
 * Note: this method is typically only called via the LOG() macro
 */
void DtlLog::logmsg(int sev, ...) {
	DtlLogConfig& config = getConfig();
	if (sev > ((level < 0) ? getDefaultLog()->level : level)) return;
	char buf[LOG_BUF_MAX];
	buf[0] = '\0';
	int len = 0;
	char* pSimpleLogBuf = NULL;  // reused pointer to basic formatting of buffer, w/o prefixes

	char csev = (sev > LOG_FINE || sev < 0) ? '?' : LOG_SEVERITY_CODES[sev];
	va_list args;
	va_start(args, sev);
	char* format = va_arg(args, char *);

	if (config.print_time) {
		time_t t;
		struct tm local_time;
		struct timeval tv;
		t = time(NULL);
		gettimeofday(&tv, NULL);
		localtime_r(&t, &local_time);
		len = SNPRINTF(buf, LOG_BUF_MAX, "[%2.2d:%2.2d:%2.2d.%3.3d] ",
					local_time.tm_hour,
					local_time.tm_min,
					local_time.tm_sec,
					(int) tv.tv_usec);
	}

	if (!config.log_func) {
		if (config.prog_name) {
			if (print_log_name) {
				len += SNPRINTF(buf+len, LOG_BUF_MAX-len, "%s[%c]: %s: ", config.prog_name, csev, log_name);
			} else {
				len += SNPRINTF(buf+len, LOG_BUF_MAX-len, "%s[%c]: ", config.prog_name, csev);
			}
		} else {
			if (print_log_name) {
				len += SNPRINTF(buf+len, LOG_BUF_MAX-len, "%s[%c]: ", log_name, csev);
			}
		}

		if (config.elapsed_time) {
			config.elapsed_time->update();
			len += SNPRINTF(buf+len, LOG_BUF_MAX-len, "<%lu> ", config.elapsed_time->getLapCpuMillis());
		}

		pSimpleLogBuf = buf+len;  // save pointer past prefixes
		len += VSNPRINTF(buf+len, LOG_BUF_MAX-len, format, args);
		strcat(buf+len, "\n");

		fwrite(buf, sizeof(char), len+1, config.log_stream);
		fflush(config.log_stream);

	} else {  // use a client supplied log_func (such as syslog_writer)
		pSimpleLogBuf = buf; // save pointer (no prefixes)
		len = VSNPRINTF(buf, LOG_BUF_MAX, format, args);
		(*config.log_func)(sev, (print_log_name ? log_name : NULL), buf, len);
	}
	va_end(args);

	// Call the auxiliary logging function for INFO or higher messages
	pLogBufWriter pAuxLog = getAuxLogFunc();
	if (pAuxLog && sev <= LOG_INFO) {
		config.aux_log_func = NULL;  // avoid recursion
		buf[len] = '\0';  // remove possible newline
		(*pAuxLog)(sev, (print_log_name ? log_name : NULL), pSimpleLogBuf, len - (pSimpleLogBuf - buf));
		config.aux_log_func = pAuxLog;  // restore
	}
}

void DtlLog::hexdump(int sev, char* label, void *pAddressIn, long lSize)
{
	if (sev > ((level < 0) ? getDefaultLog()->level : level)) return;

	if (label) {
		logmsg(sev, "hexdump of %s (%d bytes):", label, lSize);
	}

	char szBuf[100];
	long lIndent = 1;
	long lOutLen, lIndex, lIndex2, lOutLen2;
	long lRelPos;
	struct { char *pData; unsigned long lSize; } buf;
	unsigned char *pTmp,ucTmp;
	unsigned char *pAddress = (unsigned char *)pAddressIn;
	
	buf.pData   = (char *)pAddress;
	buf.lSize   = lSize; 
	
	while (buf.lSize > 0) {
		pTmp     = (unsigned char *)buf.pData;
		lOutLen  = (int)buf.lSize;
		if (lOutLen > 16) lOutLen = 16;
	
		// create a 64-character formatted output line:
		sprintf(szBuf, " >                            "
					 "                      "
#if defined(__MVS__) && !defined(_LP64)
					 "    %08X %08lX", (unsigned int)(pTmp-pAddress), (unsigned long int)pTmp);
#else
					 "    %08X %8p", (unsigned int)(pTmp-pAddress), pTmp);
#endif
		lOutLen2 = lOutLen;
	
		for(lIndex = 1+lIndent, lIndex2 = 53-15+lIndent, lRelPos = 0;
			lOutLen2;
			lOutLen2--, lIndex += 2, lIndex2++
			)
		{
			ucTmp = *pTmp++;
	
			sprintf(szBuf + lIndex, "%02X ", (unsigned short)ucTmp);
			if(!isprint(ucTmp))  ucTmp = '.'; // nonprintable char
			szBuf[lIndex2] = ucTmp;
	
			if (!(++lRelPos & 3))     // extra blank after 4 bytes
				{  lIndex++; szBuf[lIndex+2] = ' '; }
		}
	
		if (!(lRelPos & 3)) lIndex--;
	
		szBuf[lIndex  ]   = '<';
		szBuf[lIndex+1]   = ' ';
	
		logmsg(sev, "%s", szBuf);
	
		buf.pData   += lOutLen;
		buf.lSize   -= lOutLen;
	}
}

/**
 * A logging function that can be used with setLogFunc() which
 * logs to syslog.
 */
void DtlLog::syslog_writer(int sev, const char* log_name, char* logmsg, size_t len)  {

	if (sev > LOG_DEBUG) sev = LOG_DEBUG;
	if (log_name) {
		syslog(sev, "%s: %s", log_name, logmsg);
	} else {
		syslog(sev, "%s", logmsg);
	}
}


