/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */
 
#ifndef _DTL_LOG_H
#define _DTL_LOG_H 1
#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif

#ifdef __MVS__
	#define _POSIX_SOURCE
	#define _XOPEN_SOURCE 500
	#define _XOPEN_SOURCE_EXTENDED 1
	/* to get snprintf() - requires R 1.6 */
	#define _ISOC99_SOURCE
#elif _AIX
	#define _ALL_SOURCE 1
	#define _XOPEN_SOURCE_EXTENDED 1
	/* #undef HAVE_GETLINE removed to use >= AIX 7.1 getline() */
#elif defined(_STDC_C99)
	#define _XOPEN_SOURCE 600
#endif

#define _OPEN_SYS_SOCK_IPV6


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <syslog.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>


#include "ElapsedTime.h"

#ifndef LOG_TRACE
#define LOG_TRACE 8
#endif

#ifndef LOG_FINE
#define LOG_FINE 9
#endif

/* From <syslog.h>, for reference
#define LOG_EMERG       0       //A panic condition
#define LOG_ALERT       1       //Requires immediate action
#define LOG_CRIT        2       //Critical condition
#define LOG_ERR			3		//An error message
#define LOG_WARNING     4       //A warning message
#define LOG_NOTICE      5       //Requires special handling
#define LOG_INFO        6       //general info message
#define LOG_DEBUG       7       //program debug message
*/

/* add alias for badly named LOG_ERR */
#ifndef LOG_ERROR
#define LOG_ERROR LOG_ERR
#endif

/*
 * If (v)snprintf is available, use it.  Otherwise announce overflows
 * if they occur.  In either case, return the number of bytes
 * effectively written to the provided buffer.  Note that is different
 * behavior than standard (v)snprintf which returns the number of
 * bytes that WOULD have been written if there was enough room.
 * This way, the count can be used unmodified in subsequent calls to
 * SNPRINTF.
 */
#ifdef HAVE_SNPRINTF
#define SNPRINTF(buf, n, ... ) \
	({ \
		int __bw = ::snprintf((buf), (n), __VA_ARGS__ ); \
		(__bw < (n)) ? __bw : (n-1); \
	})
#else
#define SNPRINTF(buf, n, ... ) \
	({ \
		int __bw = ::sprintf((buf), __VA_ARGS__ ); \
		if (__bw > (n) ) { \
			::fprintf(stderr, "sprintf() message buffer overflow: %s(%d)", __FILE__, __LINE__); \
		} \
		__bw; \
	})
#endif

#ifdef HAVE_VSNPRINTF
#define VSNPRINTF(buf, n, format, args) \
	({ \
		int __bw = ::vsnprintf((buf), (n), format, args); \
		(__bw < (n)) ? __bw : (n-1); \
	})
#else
#define VSNPRINTF(buf, n, format, args) \
	({ \
		int __bw = ::vsprintf((buf), format, args); \
		if (__bw > (n) ) { \
			::fprintf(stderr, "vsprintf() message buffer overflow: %s(%d)", __FILE__, __LINE__); \
		} \
		__bw; \
	})
#endif

typedef void(*pLogBufWriter)(int sev, const char* log_name, char* logmsg, size_t len);

class DtlLog;

class LogTableEntry {
public:
	LogTableEntry() : name(NULL), log(NULL) {}
	char* name;
	DtlLog* log;
};

class DtlLogConfig {
public:
	static const int LOG_TABLE_MAX_SIZE;
	DtlLogConfig();

	ElapsedTime* elapsed_time;
	pLogBufWriter log_func;
	pLogBufWriter aux_log_func;
	bool call_aux_log_func;
	FILE* log_stream;
	LogTableEntry* log_table;
	int log_table_size;
	bool log_to_syslog;
	bool print_time;
	char* prog_name;
private:
	DtlLogConfig(const DtlLogConfig&);
	DtlLogConfig& operator=(const DtlLogConfig&);
};

/**
 * A logging facility, consisting of a DtlLog class and LOG macro.
 * Compile units may use a default DtlLog or declare their own with
 * distince logging levels.
 *
* To declare the use of a non-default DtlLog in a compile unit:

	// All files using the log, before including "DtlLog.h":
  	#define DTL_LOG MyLog
	#include DtlLog.h
	...
	// in a single .C compile unit, define the (extern global) DtlLog:
  	DtlLog MyLog("MyLog");
*/
class DtlLog {
public:
	static const int LOG_BUF_MAX;
	DtlLog(const char* _log_name, int init_level=-1);  // -1 -> delegate level to default logger
	~DtlLog();

	inline void setLevel(int sev) { level = sev; }
	inline int  getLevel() { return (level < 0) ? getDefaultLog()->level : level; }
	void setLevel(const char* l);
	void printLogName(bool v) { print_log_name = (v && log_name != NULL); }
	void logmsg(int sev, ...);
	void hexdump(int sev, char* label, void *pAddressIn, long lSize);

	static DtlLog* getDefaultLog() { return _DEFAULT_LOG; }
	static DtlLog& getDefaultLogRef() { return *_DEFAULT_LOG; }

	static DtlLogConfig& getConfig() {
		static DtlLogConfig config;
		return config;
	}
	static DtlLog* getLogNamed(const char* _log_name);  // returns a new log if not found

	static bool hasLogStream() {
		return getConfig().log_stream && getConfig().log_stream != stderr;
	}
	static FILE* getLogStream() {
		return getConfig().log_stream;
	}
	static void setLogStream(FILE* _log_stream) {
		if (hasLogStream()) fclose(getConfig().log_stream);
		getConfig().log_stream = _log_stream;
	}

	static const char* getProgName() { return getConfig().prog_name; }
	static void setProgName(const char* _prog_name) {
		 // makes a copy of argument string
		if (getConfig().prog_name) delete[] getConfig().prog_name;
		getConfig().prog_name = new char[strlen(_prog_name) + 1];
		strcpy(getConfig().prog_name, _prog_name);

		// if logging to syslog, then reopen with the new _PROG_NAME
		if (getConfig().log_to_syslog) {
			handleLogOption("s");
		}
	}

	static void handleLogOption(const char* opt);

	static ElapsedTime* getElapsedTime() {return getConfig().elapsed_time; }  // null if not set
	static void setPrintTime(bool v) {
		getConfig().print_time = v;
	}
	static void setPrintElapsedTime(bool v) {
		if (v) {
			if (!getConfig().elapsed_time) {
				getConfig().elapsed_time = new ElapsedTime();
			}
		} else {
			if (getConfig().elapsed_time) {
				delete getConfig().elapsed_time;
				getConfig().elapsed_time = NULL;
			}
		}
	}

	static void syslog_writer(int sev, const char* log_name, char* logmsg, size_t len);
	static pLogBufWriter getLogFunc() { return getConfig().log_func; }
	static void setLogFunc(pLogBufWriter _log_buf_writer) { getConfig().log_func = _log_buf_writer; }
	
	static pLogBufWriter getAuxLogFunc() { 
	    if (getConfig().call_aux_log_func) { 
	       return getConfig().aux_log_func; 
	    } else { 
	       return NULL; 
	    } 
	}
	
	static void setAuxLogFunc(pLogBufWriter _log_buf_writer) { 
	    getConfig().aux_log_func = _log_buf_writer; 
	    getConfig().call_aux_log_func=true;
	}

	/*Aux logging may be disabled for an info log message by doing the following:
	 * 	xxxx_LOG.enableAuxLogFunc(false);
	 *  LOG(LOG_INFO,"xxxxxxxxxxxxxxx", ...);
	 *  xxxx_LOG.enableAuxLogFunc(true);
	 */
	static void enableAuxLogFunc(bool b) { getConfig().call_aux_log_func=b;}


private:
	DtlLog();				// Used internally for default log
	int	   level;    		// LOG_* threshold
	char*  log_name;       	// prints in log messages
	bool   print_log_name;
	static DtlLog* _DEFAULT_LOG;
};

#if defined(DTL_LOG) 
	extern DtlLog DTL_LOG;
#else
	#define DTL_LOG DtlLog::getDefaultLogRef()
#endif
#define _DO_LOG(sev, ...) DTL_LOG.logmsg(sev, __VA_ARGS__)
#define _DO_HEXDUMP(sev, label, addr, len) DTL_LOG.hexdump(sev, label, addr, len)
#define LOG_ACTIVE(sev) (sev <= DTL_LOG.getLevel())

#define LOG(sev, ...) if (LOG_ACTIVE(sev)) _DO_LOG(sev, __VA_ARGS__)

#define HEXDUMP(sev, label, addr, len) if (LOG_ACTIVE(sev)) _DO_HEXDUMP(sev, label, addr, len)

/*  (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace, Fine) */
#define LOG_SEVERITY_CODES "MACEWNIDTF"


#define TRACE_CTX_BUFLEN 512

#define TRACE_CTX_DEBUG(fnName) TraceContext __traceContext(DTL_LOG, fnName, LOG_DEBUG)

#define TRACE_CTX(fnName) TraceContext __traceContext(DTL_LOG, fnName, LOG_TRACE)

#define TRACE_CTX_FINE(fnName) TraceContext __traceContext(DTL_LOG, fnName, LOG_FINE)

#define TRACE_CTX_ARGS_DEBUG(fnName, ... ) \
	char __fs[TRACE_CTX_BUFLEN]; \
	if (LOG_ACTIVE(LOG_DEBUG)) { \
		SNPRINTF(__fs, TRACE_CTX_BUFLEN, __VA_ARGS__ ); \
	} \
	TraceContext __traceContext(DTL_LOG, fnName, __fs, LOG_DEBUG)

#define TRACE_CTX_ARGS(fnName, ... ) \
	char __fs[TRACE_CTX_BUFLEN]; \
	if (LOG_ACTIVE(LOG_TRACE)) { \
		SNPRINTF(__fs, TRACE_CTX_BUFLEN, __VA_ARGS__ ); \
	} \
	TraceContext __traceContext(DTL_LOG, fnName, __fs)

#define TRACE_CTX_ARGS_FINE(fnName, ... ) \
	char __fs[TRACE_CTX_BUFLEN]; \
	if (LOG_ACTIVE(LOG_FINE)) { \
		SNPRINTF(__fs, TRACE_CTX_BUFLEN, __VA_ARGS__ ); \
	} \
	TraceContext __traceContext(DTL_LOG, fnName, __fs, LOG_FINE)

#define TRACE_CTX_FN __traceContext.fnName

#define TRACE_CTX_RETVAL( ... ) \
	if (__traceContext.isActive()) { \
		if (__traceContext.retval) delete[] __traceContext.retval; \
		__traceContext.retval = new char[TRACE_CTX_BUFLEN]; 					\
		SNPRINTF(__traceContext.retval, TRACE_CTX_BUFLEN, __VA_ARGS__ ); \
	}

#define TRACE( msg ) \
	if (__traceContext.isActive()) { \
		__traceContext.theLog.logmsg(__traceContext.level, "   %s: %s", TRACE_CTX_FN, msg); \
	}

#define TRACEV( fmt, ... ) \
	if (__traceContext.isActive()) { \
		char __buf[TRACE_CTX_BUFLEN]; \
		SNPRINTF(__buf, TRACE_CTX_BUFLEN, "   %s: %s", TRACE_CTX_FN, fmt); \
		__traceContext.theLog.logmsg(__traceContext.level, __buf, __VA_ARGS__); \
	}


class TraceContext {
public:
	DtlLog& theLog;
	const char* fnName;
	char* retval;
	int level;
	TraceContext(DtlLog& log, const char* fnName, int _level=LOG_TRACE) : theLog(log), fnName(fnName), retval(NULL), level(_level) {
		if (isActive()) {
			theLog.logmsg(level, "-> %s()", fnName);
		}
	}
	TraceContext(DtlLog& log, const char* fnName, const char* msg, int _level=LOG_TRACE) : theLog(log), fnName(fnName), retval(NULL), level(_level) {
		if (isActive()) {
			theLog.logmsg(level, "-> %s(%s)", fnName, msg);
		}
	}
	~TraceContext() {
		if (isActive()) {
			theLog.logmsg(level, "<- %s(%s)", fnName, retval ? retval : "");
		}
		if (retval) delete[] retval;
	}
	bool isActive() {
		return level <= theLog.getLevel();
	}
};


#endif /* _DTL_LOG_H */
