/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#ifndef _PTHREAD_UTIL_H
#define _PTHREAD_UTIL_H

#include "DtlLog.h"

#include <pthread.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "CoZConstants.h"

class PthreadUtil {
	private:
		static void error(const char* file, int line, ...) {
			char * format;
			char buf[2048];
			buf[0] = '\0';
			va_list args;
			va_start(args, line);
			format = va_arg(args, char *);
			SNPRINTF(buf, sizeof(buf), "%s(%d) - ", file, line);
			VSNPRINTF(buf+strlen(buf), sizeof(buf)-strlen(buf), format, args);
			va_end(args);
			if (errno) {
				LOG(LOG_ERR, "%s - [%d] %s", buf, errno, strerror(errno));
			} else {
				LOG(LOG_ERR, "%s", buf);
			}
			throw RC_SYSTEM_ERROR;
		}
			
	public:
		static void pthread_cancel(pthread_t thread, const char* file, int line) {
			errno = 0;
			if ((::pthread_cancel(thread) == -1)) error(file, line, "pthread_cancel(%d)", thread);
		}
		
		static void pthread_create(pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void *), void *arg, const char* file, int line) {
			errno = 0;
			if ((::pthread_create(thread, attr, start_routine, arg) == -1)) error(file, line, "pthread_create");
		}
		
		static void pthread_join(pthread_t thread, void **status, const char* file, int line) {
			errno = 0;
			if ((::pthread_join(thread, status)) == -1) error(file, line, "pthread_join(%d)", thread);
		}

		static void pthread_mutex_init(pthread_mutex_t* mutex, pthread_mutexattr_t *attr, const char* file, int line) {
			errno = 0;
			if ((::pthread_mutex_init(mutex, attr)) == -1) error(file, line, "pthread_mutex_init(%p, %p)", mutex, attr);
		}
		static void pthread_mutex_lock(pthread_mutex_t* mutex, const char* file, int line) {
			errno = 0;
			if ((::pthread_mutex_lock(mutex)) == -1) error(file, line, "pthread_mutex_lock(%p)", mutex);
		}
		static void pthread_mutex_unlock(pthread_mutex_t* mutex, const char* file, int line) {
			errno = 0;
			if ((::pthread_mutex_unlock(mutex)) == -1) error(file, line, "pthread_mutex_unlock(%p)", mutex);
		}
		static void pthread_cond_init(pthread_cond_t* cond, pthread_condattr_t *attr, const char* file, int line) {
			errno = 0;
			if ((::pthread_cond_init(cond, attr)) == -1) error(file, line, "pthread_cond_init(%p, %p)", cond, attr);
		}
		static void pthread_cond_wait(pthread_cond_t* cond, pthread_mutex_t* mutex, const char* file, int line) {
			errno = 0;
			if ((::pthread_cond_wait(cond, mutex)) == -1) error(file, line, "pthread_cond_wait(%p, %p)", cond, mutex);
		}
		static void pthread_cond_timedwait(pthread_cond_t* cond, pthread_mutex_t* mutex, struct timespec* t, const char* file, int line) {
			errno = 0;
			if ((::pthread_cond_timedwait(cond, mutex, t)) == -1 && errno != ETIMEDOUT && errno != EAGAIN )
				error(file, line, "pthread_cond_timedwait(%p, %p)", cond, mutex);
		}
		static void pthread_cond_signal(pthread_cond_t* cond, const char* file, int line) {
			errno = 0;
			if ((::pthread_cond_signal(cond)) == -1) error(file, line, "pthread_cond_signal(%p)", cond);
		}
};

#define Pthread_cancel(thread) (PthreadUtil::pthread_cancel(thread, __FILE__, __LINE__))
#define Pthread_create(thread, attr, start_routine, arg) (PthreadUtil::pthread_create(thread, attr, start_routine, arg, __FILE__, __LINE__))
#define Pthread_join(thread, status) (PthreadUtil::pthread_join(thread, status, __FILE__, __LINE__))
#define Pthread_mutex_init(mutex, attr) (PthreadUtil::pthread_mutex_init(mutex, attr, __FILE__, __LINE__))
#define Pthread_mutex_lock(mutex) (PthreadUtil::pthread_mutex_lock(mutex, __FILE__, __LINE__))
#define Pthread_mutex_unlock(mutex) (PthreadUtil::pthread_mutex_unlock(mutex, __FILE__, __LINE__))
#define Pthread_cond_init(cond, attr) (PthreadUtil::pthread_cond_init(cond, attr, __FILE__, __LINE__))
#define Pthread_cond_wait(cond, mutex) (PthreadUtil::pthread_cond_wait(cond, mutex, __FILE__, __LINE__))
#define Pthread_cond_timedwait(cond, mutex, t) (PthreadUtil::pthread_cond_timedwait(cond, mutex, t, __FILE__, __LINE__))
#define Pthread_cond_signal(cond) (PthreadUtil::pthread_cond_signal(cond, __FILE__, __LINE__))
#endif  /*_PTHREAD_UTIL_H*/
