/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#ifndef _SYS_UTIL_H
#define _SYS_UTIL_H

#include "DtlLog.h"

/*
 * Set generic defaults for sys_util error codes if not
 * defined earlier explicitly.
 */
static const int _SYS_UTIL_SYSTEM_ERROR = 102;
static const int _SYS_UTIL_IO_ERROR     = 103;

#ifndef SYS_UTIL_SYSTEM_ERROR
	#define SYS_UTIL_SYSTEM_ERROR _SYS_UTIL_SYSTEM_ERROR
#endif

#ifndef SYS_UTIL_IO_ERROR
	#define SYS_UTIL_IO_ERROR _SYS_UTIL_IO_ERROR
#endif

#ifndef HAVE_ISBLANK
#define isblank(ch) ((ch) == ' ' || (ch) == '\t')
#endif

#include <ctype.h>
#include <errno.h>
#define __FX
#include <fcntl.h>
#include <iconv.h>
#include <langinfo.h>
#include <locale.h>
#include <libgen.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <arpa/inet.h>

#include <sys/msg.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h> 

#define Accept(socket, sockaddr, len) (SysUtil::accept(socket, sockaddr, len, __FILE__, __LINE__))
#define Bind(socket, sockaddr, len) (SysUtil::bind(socket, sockaddr, len, __FILE__, __LINE__))
#define Close(fd) (SysUtil::close(fd, __FILE__, __LINE__))
#define Dup2(fd1, fd2) (SysUtil::dup2(fd1, fd2, __FILE__, __LINE__))
#define Fclose(stream) (SysUtil::fclose(stream, __FILE__, __LINE__))
#define Fcntl(fd, cmd, ... ) (SysUtil::fcntl(fd, cmd, __VA_ARGS__, __FILE__, __LINE__))
#define Fdopen(fd, options) (SysUtil::fdopen(fd, options, __FILE__, __LINE__))
#define Fgets(buf, n, stream) (SysUtil::fgets(buf, n, stream, __FILE__, __LINE__))
#define Fork() (SysUtil::fork(__FILE__, __LINE__))
#define Fopen(name, mode) (SysUtil::fopen(name, mode, __FILE__, __LINE__))
#define Fprintf(stream, ...) (SysUtil::fprintf(__FILE__, __LINE__, stream, __VA_ARGS__))
#define Fwrite(buffer, size, count, stream) (SysUtil::fwrite(buffer, size, count, stream, __FILE__, __LINE__))
#ifdef HAVE_GETADDRINFO
#define Getaddrinfo(nodename, servname, hints, res) (SysUtil::getaddrinfo(nodename, servname, hints, res, __FILE__, __LINE__))
#endif
#define Gethostname(name, namelen) (SysUtil::gethostname(name, namelen, __FILE__, __LINE__))
#define Getpwent() (SysUtil::getpwent(__FILE__, __LINE__))
#define Getpwuid(uid) (SysUtil::getpwuid(uid, __FILE__, __LINE__))
#define Iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft) (SysUtil::iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft))
#define Iconv_open(tocode, fromcode) (SysUtil::iconv_open(tocode, fromcode, __FILE__, __LINE__))
#define Listen(socket, backlog) (SysUtil::listen(socket, backlog, __FILE__, __LINE__))
#define Open(pathname, options, ... ) (SysUtil::open(pathname, options, __VA_ARGS__, __FILE__, __LINE__))
#define Pipe(fdinfo) (SysUtil::pipe(fdinfo, __FILE__, __LINE__))
#define Read(fs, buf, len) (SysUtil::read(fs, buf, len, __FILE__, __LINE__))
#define Read_fully(fs, buf, len) (SysUtil::read_fully(fs, buf, len, __FILE__, __LINE__))
#define Recv(sock, buf, len, flags) (SysUtil::recv(sock, buf, len, flags, __FILE__, __LINE__))
#define Select(nmsgsfds, readlist, writelist, exceptlist, timeout) (SysUtil::select(nmsgsfds, readlist, writelist, exceptlist, timeout, __FILE__, __LINE__))
#define Send(sock, buf, len) (SysUtil::send(sock, buf, len, __FILE__, __LINE__))
#define Send_fully(sock, buf, len) (SysUtil::send_fully(sock, buf, len, __FILE__, __LINE__))
#define Setenv(var_name, new_value, change_flag) (SysUtil::setenv(var_name, new_value, change_flag, __FILE__, __LINE__))
#define Setlocale(category, locale) (SysUtil::setlocale(category, locale, __FILE__, __LINE__))
#define Setsockopt(sock, lev, opt_nm, opt_val, opt_len) (SysUtil::setsockopt(sock, lev, opt_nm, opt_val, opt_len, __FILE__, __LINE__))
#define Stringcopy(dest, src, dest_len) (SysUtil::stringcopy(dest, src, dest_len))
#define Wait(status_ptr) (SysUtil::wait(status_ptr, __FILE__, __LINE__))
#define Waitpid(pid, status_ptr, options) (SysUtil::waitpid(pid, status_ptr, options, __FILE__, __LINE__))
#define Write(fs, buf, len) (SysUtil::write(fs, buf, len, __FILE__, __LINE__))
#define Write_fully(fs, buf, len) (SysUtil::write_fully(fs, buf, len, __FILE__, __LINE__))

class SysUtil {
	private:
		static void err_impl(const char* file, int line, int rc, char* format, va_list args) {
			char buf[2048];
			buf[0] = '\0';
			SNPRINTF(buf, sizeof(buf), "%s(%d) - ", file, line);
			VSNPRINTF(buf+strlen(buf), sizeof(buf)-strlen(buf), format, args);
			if (errno) {
				LOG(LOG_ERR, "%s - [%d] %s", buf, errno, strerror(errno));
			} else {
				LOG(LOG_ERR, "%s", buf);
			}
			throw rc;
		}

		static void error(const char* file, int line, ...) {
			char* format;
			va_list args;
			va_start(args, line);
			format = va_arg(args, char *);
			err_impl(file, line, SYS_UTIL_SYSTEM_ERROR, format, args);
			va_end(args);
		}

		static void ioerror(const char* file, int line, ...) {
			char* format;
			va_list args;
			va_start(args, line);
			format = va_arg(args, char *);
			err_impl(file, line, SYS_UTIL_IO_ERROR, format, args);
			va_end(args);
		}

	public:

		static int accept(int socket, struct sockaddr* sockaddr, socklen_t* len, const char* file, int line) {
			errno = 0;
			int fd;
			if ((fd = ::accept(socket, sockaddr, len)) == -1) error(file, line, "accept");
			return fd;
		}
		
		static void bind(int socket, struct sockaddr* sockaddr, socklen_t len, const char* file, int line) {
			errno = 0;
			if (::bind(socket, sockaddr, len) == -1) error(file, line, "bind");			
		}
		
		static void close(int fd, const char* file, int line) {
			errno = 0;
			if (::close(fd) == -1) error(file, line, "close");
		}

		static int dup2(int fd1, int fd2, const char* file, int line) {
			int fd;
			errno = 0;
			if ((fd = ::dup2(fd1, fd2) == -1)) error(file, line, "dup2");
 			return fd;
		}

		static FILE *fdopen(int fd, const char* options, const char* file, int line) {
			FILE* f;
			if (!(f = ::fdopen(fd, options))) error(file, line, "fdopen(%d, %s)", fd, options);
			return f;
		}

		static int fclose(FILE* stream, const char* file, int line) {
			int	n;
			errno = 0;
			if ((n = ::fclose(stream)) != 0) error(file, line, "fclose()");
			return n;
		}

		static int fcntl(int fd, int cmd, int arg, const char* file, int line) {
			int	n;
			errno = 0;
			if ((n = ::fcntl(fd, cmd, arg)) == -1) error(file, line, "fcntl(%d,%d,%h)", fd, cmd, arg);
			return n;
		}

		static int fcntl(int fd, int cmd, const char* file, int line) {
			int	n;
			errno = 0;
			if ((n = ::fcntl(fd, cmd)) == -1) error(file, line, "fcntl(%d,%d)", fd, cmd);
			return n;
		}

		static char* fgets(char* buf, int n, FILE* stream, const char* file, int line) {
			errno = 0;
			char* p = ::fgets(buf, n, stream);
			if (!p && ferror(stream)) ioerror(file, line, "fgets(%p,%d,%p)", buf, n, stream);
			return p;
		}

		static FILE* fopen(const char* filename, const char* mode, const char* file, int line) {
			FILE* f;
			errno = 0;
			if (!(f = ::fopen(filename, mode))) error(file, line, "fopen(%s,%s)", filename, mode);
			return f;
		}

		static pid_t fork(const char* file, int line) {
			pid_t pid;
			errno = 0;
#ifdef __ZOS__
			// Attempt to fork(), retrying after EAGAIN failures for up to a total of 5.
			// This code is in response to behavior described by USS DOC APAR OA30943 (Problem #1). 
			int i = 0;
			for (i=0; i<5 && ((pid = ::fork()) == -1); i++) {
				if (errno == EAGAIN)
					continue;
			}
			if (pid == -1) error(file, line, "fork (zos)");
#else
			if ((pid = ::fork()) == -1) error(file, line, "fork");
#endif
			return pid;
		}

		static int fprintf(const char* file, int line, FILE* stream, ...) {
			int n;
			errno = 0;
			char* format;
			va_list args;
			va_start(args, stream);
			format = va_arg(args, char *);
			if((n = ::vfprintf(stream, format, args)) < 0) ioerror(file, line, "fprintf");
			va_end(args);
			return n;
		}

		static size_t fwrite(const void * buffer, size_t size, size_t count, FILE* stream, const char* file, int line) {
			errno = 0;
			size_t nWritten = ::fwrite(buffer, size, count, stream);
			if (nWritten < count) {
				ioerror(file, line, "fwrite");
			}
			return nWritten;
		}

#ifdef HAVE_GETADDRINFO
		static void getaddrinfo(const char *nodename, const char *servname, const struct addrinfo *hints, struct addrinfo **res, const char* file, int line) {
			int rc;
			if ((rc = ::getaddrinfo(nodename, servname, hints, res))) {
				if (rc != EAI_SYSTEM) {
					errno = 0; //Unless EAI_SYSTEM, errno is ignored
				}
				ioerror(file, line, "getaddrinfo(%s,%s,...) - [%d] %s", nodename, servname, rc, gai_strerror(rc));
			}
        }
#endif

		static void gethostname(char *name, size_t namelen, const char* file, int line) {
			errno = 0;
			if (::gethostname(name, namelen) == -1) error(file, line, "gethostname");
		}

		//ICONV_CONST is set to const or "" depending on the platform.
		static size_t iconv(iconv_t cd, char** inbuf, size_t* inbytesleft, char** outbuf, size_t* outbytesleft) {
			return ::iconv(cd, (ICONV_CONST char**) inbuf, inbytesleft, outbuf, outbytesleft);
		}

		static iconv_t iconv_open(const char *tocode, const char *fromcode, const char* file, int line) {
			iconv_t cd;
			errno = 0;
			if ((cd = ::iconv_open(tocode, fromcode)) == (iconv_t) -1)
				error(file, line, "iconv_open(%s, %s)", tocode, fromcode);
			return cd;
		}
		
		static void listen(int socket, int backlog, const char* file, int line) {
			errno = 0;
			if (::listen(socket, backlog) == -1) error(file, line, "listen");			
		}
		
		static int open(const char *pathname, int options, const char* file, int line) {
			errno = 0;
			int fd;
			if ((fd = ::open(pathname, options)) == -1) error(file, line, "open(%s,%h)", pathname, options);
			return fd;
		}

		static int open(const char *pathname, int options, mode_t mode, const char* file, int line) {
			errno = 0;
			int fd;
			if ((fd = ::open(pathname, options, mode)) == -1) error(file, line, "open(%s,%h,%h)", pathname, options, mode);
			return fd;
		}

		static struct passwd* getpwent(const char* file, int line) {
			struct passwd* pw;
			errno = 0;
			if (!(pw = ::getpwent())) error(file, line, "getpwent");
			return pw;
		}

		static struct passwd* getpwuid(uid_t uid, const char* file, int line) {
			struct passwd* pw;
			errno = 0;
#ifdef __ZOS__
			char* name;
			if(!(name = __getlogin1())) {
				pw = NULL;
			} else {
				pw = getpwnam(name);
			}
			if (!pw) {  // fall back to the normal way
				if(!(pw = ::getpwuid(uid))) error(file, line, "getpwuid(%d)", uid);
			}
#else
			if(!(pw = ::getpwuid(uid))) error(file, line, "getpwuid(%d)", uid);
#endif
			return pw;
		}

		static void pipe(int fdinfo[2], const char* file, int line) {
			errno = 0;
			if (::pipe(fdinfo) == -1) error(file, line, "pipe");
		}

		static ssize_t read(int fs, void* buf, ssize_t len, const char* file, int line) {
			ssize_t n;
			errno = 0;
			if ((n = ::read(fs, buf, len)) == -1) ioerror(file, line, "read(%d, %p, %d)", fs, buf, len);
			return n;
		}

		static ssize_t recv(int sock, void* buf, ssize_t len, int flags, const char* file, int line) {
			ssize_t n;
			errno = 0;
			if ((n = ::recv(sock, buf, len, flags)) == -1) ioerror(file, line, "recv(%d, %p, %d)", sock, buf, len);
			return n;
		}

		// NB: this should not be used on a O_NONBLOCK file/socket/pipe
		static size_t read_fully(int fd, void* buf, size_t len, const char* file, int line) {
			char* pBuf = (char*)buf;
			size_t nextLen;

			while ((nextLen = len - (pBuf - (char*)buf)) >  0) {
				errno = 0;
				ssize_t nRead = ::read(fd, pBuf, nextLen);
				if (nRead == 0) {
					if (pBuf > (char*)buf) {
						// can't read partially before getting EOF
						ioerror(file, line, "read_fully(%d, %p, %d) unexpected EOF", fd, buf, len);
					} else {
						// but getting nothing but an EOF returns 0
						return 0; // EOF
					}
				} else if (nRead > 0) {
					pBuf += nRead;
				} else {
					if (errno == EINTR) {
						continue;   // try again if interrupted by a signal
					} else {
						ioerror(file, line, "read_fully(%d, %p, %d)", fd, buf, len);
					}
				}
			}
			return len;  // return the request len
		}

		static int select(int nmsgsfds, fd_set *readlist, fd_set *writelist, fd_set *exceptlist, struct timeval *timeout, const char* file, int line) {
			int n;
			errno = 0;
			if ((n = ::select(nmsgsfds, readlist, writelist, exceptlist, timeout)) == -1)
				error(file, line, "select");
			return n;
		}

		static void setenv(const char *var_name, const char *new_value, int change_flag, const char* file, int line) {
			errno = 0;
			#ifdef HAVE_SETENV
				if (::setenv(var_name, new_value, change_flag) != 0)
					error(file, line, "setenv(%s,%s)", var_name, new_value);
			#else
				char* new_var = new char[strlen(var_name) + strlen(new_value) + 2];
				sprintf(new_var, "%s=%s", var_name, new_value);
				if (::putenv(new_var) != 0)
					error(file, line, "putenv(%s)", new_var);
			#endif
		}

		/* Set locale from the environment.
		 * Note - if NLS is not setup properly, this call will return NULL.
	 	 * In this case, issue a warning an continue.
	 	 */
		static char* setlocale(int category, const char *locale, const char* file, int line) {
			char* l;
			errno = 0;
			if (!(l = ::setlocale(category, locale)))
				LOG(LOG_WARNING,"setlocale() returned NULL.  The NLS environment may not be properly configured");
			return l;
		}

		static void setsockopt(int socket, int level, int option_name, const void *option_value, socklen_t option_length, const char* file, int line) {
			errno = 0;
			if (::setsockopt(socket, level, option_name, option_value, option_length) == -1)
				ioerror(file, line, "setsockopt");
		}

		// String copy function similar to strncpy, but always null terminates.  At most
		// dest_len-1 characters are copied from src.
		static char* stringcopy(char* dest, const char* src, size_t dest_len) {
			int i;
			if (dest_len == 0) return dest;
			strncpy(dest, src, dest_len-1);
			*(dest+dest_len-1) = '\0';
			return dest;
		}

		/*
		 * Call wait() in a loop, redriving on EINTR.  If ECHILD is received, return -1 for
		 * caller to handle, otherwise return the pid.
		 */
		static pid_t wait(int *status_ptr, const char* file, int line) {
			pid_t p = -1;
			errno = 0;
			while (p == -1) {
				if ((p = ::wait(status_ptr)) == -1) {
					if (errno == EINTR) continue;
					if (errno == ECHILD) {
						LOG(LOG_DEBUG,"wait() returned ECHILD.");
						break;
					}
					error(file, line, "wait");
				}
			}
   			return p;
		}

		/*
		 * Call waitpid in a loop, redriving on EINTR.  If ECHILD is received, return -1 for
		 * caller to handle, otherwise return the pid.
		 */
		static pid_t waitpid(pid_t pid, int *status_ptr, int options, const char* file, int line) {
			pid_t p = -1;
			errno = 0;
			while (p == -1) {
				if ((p = ::waitpid(pid, status_ptr, options)) == -1) {
					if (errno == EINTR) continue;
					if (errno == ECHILD) {
						LOG(LOG_DEBUG,"waitpid(%d) returned ECHILD.", pid);
						break;
					}
					error(file, line, "waitpid");
				}
    		}
   			return p;
		}

		static ssize_t write(int fs, const void* buf, ssize_t len, const char* file, int line) {
			ssize_t n;
			errno = 0;
			if ((n = ::write(fs, buf, len)) == -1) ioerror(file, line, "write(%d, %p, %d)", fs, buf, len);
			return n;
		}

		static ssize_t send(int sock, const void* buf, ssize_t len, const char* file, int line) {
			ssize_t n;
			errno = 0;
			if ((n = ::send(sock, buf, len, 0)) == -1) ioerror(file, line, "sock(%d, %p, %d)", sock, buf, len);
			return n;
		}

		static void write_fully(int fd, const void* buf, size_t len, const char* file, int line) {
			const char* pBuf = (const char*)buf;
			size_t nextLen = len;

			while (true) {
				errno = 0;
				ssize_t nWritten = ::write(fd, pBuf, nextLen);
				if (nWritten == -1) {
					if (errno == EAGAIN) continue;
					ioerror(file, line, "write_fully(%d, %p, %d)", fd, buf, len);
				} else if (nWritten == nextLen) {
					return;  // complete

				} else {  // a short write
					pBuf += nWritten;
					nextLen -= nWritten;
				}
			}
		}

		static void send_fully(int sock, const void* buf, size_t len, const char* file, int line) {
			const char* pBuf = (const char*)buf;
			size_t nextLen = len;

			while (true) {
				errno = 0;
				ssize_t nWritten = ::send(sock, pBuf, nextLen, 0);
				if (nWritten == -1) {
					if (errno == EAGAIN) continue;
					ioerror(file, line, "send_fully(%d, %p, %d)", sock, buf, len);
				} else if (nWritten == nextLen) {
					return;  // complete

				} else {  // a short write
					pBuf += nWritten;
					nextLen -= nWritten;
					/* traces of this indicate that we don't seem to be doing any short sends....
					   LOG(LOG_FINE, "send_fully(sock=%d, len=%l) sent %lu bytes",
							sock, nextLen, nWritten); */
				}
			}
		}
};

#endif  /*_SYS_UTIL_H*/
